"""Utilities for exploratory data analysis (EDA)."""

from __future__ import annotations

from mismo.eda._plot import distribution_chart as distribution_chart
from mismo.eda._plot import distribution_dashboard as distribution_dashboard
