from __future__ import annotations

from mismo.linkage._combine import difference as difference
from mismo.linkage._combine import intersect as intersect
from mismo.linkage._combine import (
    unify_links_min_intersection as unify_links_min_intersection,
)
from mismo.linkage._combine import union as union
from mismo.linkage._linkage import Linkage as Linkage
from mismo.linkage._linkage import filter_links as filter_links
from mismo.linkage._sample import sample_all_links as sample_all_links
