"""Utils for cleaning and comparing email addresses."""

from __future__ import annotations

from mismo.lib.email._core import EmailMatchLevel as EmailMatchLevel
from mismo.lib.email._core import EmailsDimension as EmailsDimension
from mismo.lib.email._core import ParsedEmail as ParsedEmail
from mismo.lib.email._core import clean_email as clean_email
from mismo.lib.email._core import match_level as match_level
