"""Houses domain-specific stuff, such as for names, addresses, locations, etc."""

from __future__ import annotations

from mismo.lib import email as email
from mismo.lib import geo as geo
from mismo.lib import name as name
from mismo.lib import phone as phone
